@@ -37,6 +37,7 @@
 #include <qfileinfo.h>
 #include <qregexp.h>
 #include <qstring.h>
+#include <qdir.h>
 
 #include <iostream>
 #include <ctype.h>
@@ -139,7 +140,6 @@ GDBController::GDBController(VariableTree *varTree, FramestackWidget *frameStack
         badCore_(QString()),
         state_(s_dbgNotStarted|s_appNotStarted|s_silent),
         programHasExited_(false),
-        backtraceDueToProgramStop_(false),
         dom(projectDom),
         config_breakOnLoadingLibrary_(true),
         config_forceBPSet_(true),
@@ -208,7 +208,6 @@ void GDBController::configure()
         bool restart = false;
         if (stateIsOn(s_appBusy))
         {
-            setStateOn(s_silent);
             pauseApp();
             restart = true;
         }
@@ -375,6 +374,8 @@ void GDBController::removeInfoRequests()
 // commands as well.
 void GDBController::pauseApp()
 {
+    setStateOn(s_silent);
+
     int i = cmdList_.count();
     while (i)
     {
@@ -408,22 +409,17 @@ void GDBController::actOnProgramPause(const QString &msg)
         // and we must reset the active flag
         viewedThread_ = -1;
         currentFrame_ = 0;
-        varTree_->setActiveFlag();
-        backtraceDueToProgramStop_ = true;
 
         // These two need to be actioned immediately. The order _is_ important
         if (stateIsOn(s_viewThreads))
             queueCmd(new GDBCommand("info thread", NOTRUNCMD, INFOCMD, INFOTHREAD), true);
 
         queueCmd(new GDBCommand("backtrace", NOTRUNCMD, INFOCMD, BACKTRACE), true);
-        if (stateIsOn(s_viewLocals))
-        {
-            queueCmd(new GDBCommand("info args", NOTRUNCMD, INFOCMD, ARGS));
-            queueCmd(new GDBCommand("info local", NOTRUNCMD, INFOCMD, LOCALS));
-        }
 
-        varTree_->findWatch()->requestWatchVars();
-        varTree_->findWatch()->setActive();
+        // The above should have changed viewedThread_, if we're in MT
+        // program. 
+        emit currentFrame(currentFrame_, viewedThread_);
+
         emit acceptPendingBPs();
     }
 }
@@ -443,13 +439,12 @@ void GDBController::programNoApp(const QString &msg, bool msgBox)
     // and we must reset the active flag
     viewedThread_ = -1;
     currentFrame_ = 0;
-    varTree_->setActiveFlag();
 
-    // Now wipe the tree out
-    varTree_->viewport()->setUpdatesEnabled(false);
-    varTree_->trim();
-    varTree_->viewport()->setUpdatesEnabled(true);
-    varTree_->repaint();
+    // Delete the tty that waits for application output. Without
+    // this, KDevelop will be hanging, eating 100% CPU. Don't know
+    // why and don't understand this TTY stuff either.
+    delete tty_;
+    tty_ = 0;
 
     frameStack_->clear();
 
@@ -511,7 +506,16 @@ void GDBController::parseLine(char* buf)
             // application is running.
             // And the user does this to stop the program also.
             if (strstr(buf+23, "SIGINT") && stateIsOn(s_silent))
+            {
+                // If this is explicit break into the program, show
+                // the source line
+                if (stateIsOn(s_explicitBreakInto))
+                {
+                    setStateOff(s_silent|s_explicitBreakInto);
+                    actOnProgramPause(QString(buf));                
+                }
                 return;
+            }
 
             // Whenever we have a signal raised then tell the user, but don't
             // end the program as we want to allow the user to look at why the
@@ -615,11 +619,38 @@ void GDBController::parseLine(char* buf)
     if (strncmp(buf, "Brea", 4) == 0 ||
         strncmp(buf, "Hard", 4) == 0)
     {
-        // Starts with "Brea" so assume "Breakpoint" and just get a full
-        // breakpoint list. Note that the state is unchanged.
-        // Much later: I forget why I did it like this :-o
-        queueCmd(new GDBCommand("info breakpoints",
-                                        NOTRUNCMD, NOTINFOCMD, BPLIST));
+
+        const Breakpoint* traced = 0;
+        if (strncmp(buf, "Brea", 4) == 0)
+        {
+            char* ptr = buf + strlen("Breakpoint ");
+            int id = atoi(ptr);
+
+            if (tracedBreakpoints_.contains(id))
+            {
+                traced = tracedBreakpoints_[id];
+            }
+        }
+
+        if (traced)
+        {
+            // Prevent showing current file position in UI.
+            setStateOn(s_silent); 
+
+            // Marking this as "NOTINFOCMD". Otherwise, "continue" will
+            // immediately remove "printf" from the queue.
+            queueCmd(new GDBCommand(("printf " + traced->traceRealFormatString()).latin1(),
+                                        NOTRUNCMD, NOTINFOCMD, TRACING_PRINTF));
+            queueCmd(new GDBCommand("continue", RUNCMD, NOTINFOCMD, 0));
+        }
+        else
+        {
+            // Starts with "Brea" so assume "Breakpoint" and just get a full
+            // breakpoint list. Note that the state is unchanged.
+            // Much later: I forget why I did it like this :-o
+            queueCmd(new GDBCommand("info breakpoints",
+                                    NOTRUNCMD, NOTINFOCMD, BPLIST));
+        }
 
 //        kdDebug(9012) << "Parsed (BP) <" << buf << ">" << endl;
         return;
@@ -818,15 +849,6 @@ void GDBController::parseProgramLocation(char *buf)
 void GDBController::parseBacktraceList(char *buf)
 {
     frameStack_->parseGDBBacktraceList(buf);
-    if (backtraceDueToProgramStop_)
-    {
-        varTree_->trimExcessFrames();
-        VarFrameRoot *frame = varTree_->findFrame(currentFrame_, viewedThread_);
-        if (frame)
-            frame->setFrameName(
-                frameStack_->getFrameName(currentFrame_, viewedThread_));
-        backtraceDueToProgramStop_ = false;
-    }
 }
 
 // **************************************************************************
@@ -835,7 +857,6 @@ void GDBController::parseThreadList(char *buf)
 {
     frameStack_->parseGDBThreadList(buf);
     viewedThread_ = frameStack_->viewedThread();
-    varTree_->setCurrentThread(viewedThread_);
 }
 
 // **************************************************************************
@@ -846,11 +867,20 @@ void GDBController::parseBreakpointSet(char *buf)
 {
     if (GDBSetBreakpointCommand *BPCmd = dynamic_cast<GDBSetBreakpointCommand*>(currentCmd_))
     {
+        const Breakpoint* bp = BPCmd->breakpoint();
+
         // ... except in this case :-) A -1 key tells us that this is
         // a special internal breakpoint, and we shouldn't do anything
         // with it. Currently there are _no_ internal breakpoints.
-        if (BPCmd->getKey() != -1)
-            emit rawGDBBreakpointSet(buf, BPCmd->getKey());
+        if (bp->key() != -1)
+        {            
+            emit rawGDBBreakpointSet(buf, bp->key());
+
+            // Now, gdb id of breakpoint should be set. If it's 
+            // tracepoint, remember it.
+            if (bp->tracingEnabled())
+                tracedBreakpoints_[bp->dbgId()] = bp;
+        }
     }
 }
 
@@ -933,40 +963,14 @@ void GDBController::parseFrameSelected(char *buf)
 // inactive.
 void GDBController::parseLocals(char type, char *buf)
 {
-    varTree_->viewport()->setUpdatesEnabled(false);
-
-    // The locals are always attached to the currentFrame
-    VarFrameRoot *frame = varTree_->findFrame(currentFrame_, viewedThread_);
-    if (!frame)
-    {
-        frame = new VarFrameRoot(varTree_, currentFrame_, viewedThread_);
-        frame->setFrameName(
-                frameStack_->getFrameName(currentFrame_, viewedThread_));
-    }
-
-    Q_ASSERT(frame);
-
     if (type == (char) ARGS)
     {
-        frame->setParams(buf);
+        emit parametersReady(buf);
     }
     else
     {
-        frame->setLocals(buf);
-        // Trim the whole tree when we're on the top most
-        // frame so that they always see only "frame 0" on a program stop.
-        // User selects frame 1, will show both frame 0 and frame 1.
-        // Reselecting a frame 0 regenerates the data and therefore trims
-        // the whole tree _but_ all the items in every frame will be active
-        // so nothing will be deleted.
-        if (currentFrame_ == 0 || viewedThread_ == -1)
-            varTree_->trim();
-        else
-            frame->trim();
+        emit localsReady(buf);
     }
-
-    varTree_->viewport()->setUpdatesEnabled(true);
-    varTree_->repaint();
 }
 
 // **************************************************************************
@@ -1053,6 +1057,9 @@ char *GDBController::parseCmdBlock(char *buf)
         case LIBRARIES:
             emit rawGDBLibraries      (buf);
             break;
+        case TRACING_PRINTF:
+            emit tracingOutput        (buf);
+            break;
 //         case DETACH:
 //             setStateOff(s_attached);
 //             break;
@@ -1152,9 +1159,10 @@ char *GDBController::parse(char *buf)
 
 // **************************************************************************
 
-void GDBController::setBreakpoint(const QCString &BPSetCmd, int key)
+void GDBController::setBreakpoint(const QCString &BPSetCmd, 
+                                  const Breakpoint* bp)
 {
-    queueCmd(new GDBSetBreakpointCommand(BPSetCmd, key));
+    queueCmd(new GDBSetBreakpointCommand(BPSetCmd, bp));
 }
 
 // **************************************************************************
@@ -1190,6 +1198,12 @@ void GDBController::modifyBreakpoint( const Breakpoint& BP )
                             BP.isEnabled() ? "enable" : "disable",
                             BP.dbgId()), NOTRUNCMD, NOTINFOCMD));
 
+        if (BP.changedTracing())
+            if (BP.tracingEnabled())
+                tracedBreakpoints_[BP.dbgId()] = &BP;
+            else
+                tracedBreakpoints_.remove(BP.dbgId());
+
         //        BP.setDbgProcessing(true);
         // Note: this is NOT an info command, because gdb doesn't explictly tell
         // us that the breakpoint has been deleted, so if we don't have it the
@@ -1250,23 +1264,47 @@ void GDBController::slotStart(const QString& shell, const DomUtil::PairList& run
              this,        SLOT(slotDbgProcessExited(KProcess*)) );
 
     application_ = application;
+
+    QString gdb = "gdb";
+    // Prepend path to gdb, if needed. Using QDir,
+    // path can either end with slash, or not.
+    if (!config_gdbPath_.isEmpty())
+    {
+        QDir gdb_path(config_gdbPath_);
+        gdb = gdb_path.filePath("gdb");
+    }
+
     if (!shell.isEmpty())
     {
-        *dbgProcess_ << "/bin/sh" << "-c" << shell + " " +config_gdbPath_
-                      + "gdb " + application + " -fullname -nx -quiet";
-        emit gdbStdout(QString( "/bin/sh -c " + shell + " " +config_gdbPath_
-                      + "gdb " + application + " -fullname -nx -quiet" ).latin1());
+        *dbgProcess_ << "/bin/sh" << "-c" << shell + " " + gdb
+                      + " " + application + " -fullname -quiet";
+        emit gdbStdout(QString( "/bin/sh -c " + shell + " " + gdb
+                      + " " + application + " -fullname -quiet" ).latin1());
     }
     else
     {
-        *dbgProcess_ << config_gdbPath_ + "gdb" << application
-                        << "-fullname" << "-nx" << "-quiet";
-        emit gdbStdout(QString( config_gdbPath_ + "gdb " + application +
-                        " -fullname -nx -quiet" ).latin1());
+        *dbgProcess_ << gdb << application
+                        << "-fullname" << "-quiet";
+        emit gdbStdout(QString( gdb + " " + application +
+                        " -fullname -quiet" ).latin1());
     }
 
-    dbgProcess_->start( KProcess::NotifyOnExit,
-                        KProcess::Communication(KProcess::All));
+    if (!dbgProcess_->start( KProcess::NotifyOnExit,
+                             KProcess::Communication(KProcess::All)))
+    {
+        KMessageBox::error(
+            0, 
+            i18n("<b>Could not start debugger.</b>"
+                 "<p>Could not run '%1'. "
+                 "Make sure that the path name is specified correctly."
+                ).arg(dbgProcess_->args()[0]),
+            i18n("Could not start debugger"));
+
+        delete tty_;
+        tty_ = 0;
+
+        return;
+    }
 
     setStateOff(s_dbgNotStarted);
     emit dbgStatus ("", state_);
@@ -1304,6 +1342,8 @@ void GDBController::slotStart(const QString& shell, const DomUtil::PairList& run
                             NOTINFOCMD));
     queueCmd(new GDBCommand("handle SIG41 pass nostop noprint", NOTRUNCMD,
                             NOTINFOCMD));
+    queueCmd(new GDBCommand("handle SIG42 pass nostop noprint", NOTRUNCMD,
+                            NOTINFOCMD));
     queueCmd(new GDBCommand("handle SIG43 pass nostop noprint", NOTRUNCMD,
                             NOTINFOCMD));
 
@@ -1428,6 +1468,13 @@ void GDBController::slotStopDebugger()
     delete dbgProcess_;    dbgProcess_ = 0;
     delete tty_;           tty_ = 0;
 
+    // The gdb output buffer might contain start marker of some
+    // previously issued command that crashed gdb (so there's no end marker)
+    // If we don't clear this, then after restart, we'll be trying to search
+    // for the end marker of the command issued in previous gdb session,
+    // and never succeed. 
+    gdbOutputLen_ = 0;
+
     state_ = s_dbgNotStarted | s_appNotStarted | s_silent;
     emit dbgStatus (i18n("Debugger stopped"), state_);
 }
@@ -1489,7 +1536,7 @@ void GDBController::slotRun()
         if (!config_runShellScript_.isEmpty()) {
             // Special for remote debug...
             QCString tty(tty_->getSlave().latin1());
-            QCString options = QCString(" 2>&1 >") + tty + QCString(" <") + tty;
+            QCString options = QCString(">") + tty + QCString("  2>&1 <") + tty;
 
             KProcess *proc = new KProcess;
 
@@ -1513,7 +1560,25 @@ void GDBController::slotRun()
             // Note: script could contain "run" or "continue"
         }
         else {
-            queueCmd(new GDBCommand("run", RUNCMD, NOTINFOCMD, 0));
+
+            // The program is run locally. Need the make sure it's executable,
+            // just in case.
+            QFileInfo app(application_);
+            if (app.isExecutable())
+            {
+                queueCmd(new GDBCommand("run", RUNCMD, NOTINFOCMD, 0));
+            }
+            else
+            {
+                KMessageBox::error(
+                    0, 
+                    i18n("<b>Could not run application '%1'.</b>"
+                         "<p>The application does not have the executable bit set. "
+                         "Try rebuilding the project, or change permissions "
+                         "manually."
+                        ).arg(app.fileName()),
+                    i18n("Could not run application"));
+            }
         }
     }
     else {
@@ -1521,6 +1586,20 @@ void GDBController::slotRun()
     }
 }
 
+
+void GDBController::slotRestart()
+{
+    if (stateIsOn(s_dbgNotStarted|s_shuttingDown))
+        return;
+
+    if (stateIsOn(s_appBusy))
+    {
+        pauseApp();
+    }
+
+    queueCmd(new GDBCommand("run", RUNCMD, NOTINFOCMD, 0));        
+}
+
 // **************************************************************************
 
 void GDBController::slotRunUntil(const QString &fileName, int lineNum)
@@ -1539,6 +1618,19 @@ void GDBController::slotRunUntil(const QString &fileName, int lineNum)
 
 // **************************************************************************
 
+void GDBController::slotJumpTo(const QString &fileName, int lineNum)
+{
+    if (stateIsOn(s_appBusy|s_dbgNotStarted|s_shuttingDown))
+        return;
+
+    if (!fileName.isEmpty()) {
+        queueCmd(new GDBCommand(QCString().sprintf("tbreak %s:%d", fileName.latin1(), lineNum), NOTRUNCMD, NOTINFOCMD, 0));
+        queueCmd(new GDBCommand(QCString().sprintf("jump %s:%d", fileName.latin1(), lineNum), RUNCMD, NOTINFOCMD, 0));
+    }
+}
+
+// **************************************************************************
+
 void GDBController::slotStepInto()
 {
     if (stateIsOn(s_appBusy|s_appNotStarted|s_shuttingDown))
@@ -1592,6 +1684,7 @@ void GDBController::slotStepOutOff()
 // Only interrupt a running program.
 void GDBController::slotBreakInto()
 {
+    setStateOn(s_explicitBreakInto);
     pauseApp();
 }
 
@@ -1616,14 +1709,13 @@ void GDBController::slotBPState( const Breakpoint& BP )
 
         // When forcing breakpoints to be set/unset, interrupt a running app
         // and change the state.
-        setStateOn(s_silent);
         pauseApp();
         restart = true;
     }
 
     if (BP.isActionAdd())
     {
-        setBreakpoint(BP.dbgSetCommand().latin1(), BP.key());
+        setBreakpoint(BP.dbgSetCommand().latin1(), &BP);
         //        BP.setDbgProcessing(true);
     }
     else
@@ -1632,6 +1724,7 @@ void GDBController::slotBPState( const Breakpoint& BP )
         {
             clearBreakpoint(BP.dbgRemoveCommand().latin1());
             //            BP.setDbgProcessing(true);
+            tracedBreakpoints_.remove(BP.dbgId());
         }
         else
         {
@@ -1662,7 +1755,6 @@ void GDBController::slotClearAllBreakpoints()
 
         // When forcing breakpoints to be set/unset, interrupt a running app
         // and change the state.
-        setStateOn(s_silent);
         pauseApp();
         restart = true;
     }
@@ -1758,28 +1850,49 @@ void GDBController::slotSelectFrame(int frameNo, int threadNo, bool needFrames)
     viewedThread_ = threadNo;
     currentFrame_ = frameNo;
 
-    // Find or add the frame details. hold onto whether it existed because
-    // we're about to create one if it didn't.
-    VarFrameRoot *frame = varTree_->findFrame(frameNo, viewedThread_);
-    if (!frame)
-    {
-        frame = new VarFrameRoot(varTree_, currentFrame_, viewedThread_);
-        frame->setFrameName(
-                frameStack_->getFrameName(currentFrame_, viewedThread_));
-    }
+    emit currentFrame(frameNo, threadNo);
+}
 
-    Q_ASSERT(frame);
-    if (stateIsOn(s_viewLocals))
+/** Produces backtrace for the specified thread, or for main
+    thread if threadNo is -1. Does not change the current thread or
+    the current thread.
+*/
+void GDBController::slotProduceBacktrace(int threadNo)
+{
+    QString command;
+    if (threadNo != -1)
     {
-        // Have we already got these details?
-        if (frame->needLocals())
-        {
-            // Add the frame params to the variable list
-            // and ask for the locals
-            queueCmd(new GDBCommand("info args", NOTRUNCMD, INFOCMD, ARGS));
-            queueCmd(new GDBCommand("info local", NOTRUNCMD, INFOCMD, LOCALS));
-        }
+        command = QString("thread apply %1 backtrace").arg(threadNo);
     }
+    else
+    {
+        command = "backtrace";
+    }
+    queueCmd(new GDBCommand(command.local8Bit(), 
+             NOTRUNCMD, INFOCMD, BACKTRACE));    
+}
+
+void GDBController::slotProduceVariablesInfo()
+{
+    queueCmd(new GDBCommand("info args", NOTRUNCMD, INFOCMD, ARGS));
+    queueCmd(new GDBCommand("info local", NOTRUNCMD, INFOCMD, LOCALS));    
+}
+
+// **************************************************************************
+
+void GDBController::slotSetValue(const QString& expression, 
+                                 const QString& value)
+{
+    // Need explicit "set var", not just "set" because gdb is
+    // confused on
+    // 
+    //    set B::k=10
+    //
+    queueCmd(new GDBCommand(QString("set var %1=%2").arg(expression).arg(value)
+                            .local8Bit(),
+                            NOTRUNCMD,
+                            INFOCMD,
+                            SETVALUE));
 }
 
 // **************************************************************************
@@ -1792,7 +1905,7 @@ void GDBController::slotVarItemConstructed(VarItem *item)
     // jw - name and value come from "info local", for the type we
     // send a "whatis <varName>" here.
     //rgr: remove any format modifier
-    QString strName = item->fullName();
+    QString strName = item->gdbExpression();
     strName.remove( QRegExp("/[xd] ", FALSE) );
     queueCmd(new GDBItemCommand(item, QCString("whatis ") + strName.latin1(),
                                 false, WHATIS));
@@ -1810,23 +1923,15 @@ void GDBController::slotExpandItem(TrimmableItem *genericItem)
     VarItem *varItem;
     if ((varItem = dynamic_cast<VarItem*>(genericItem)))
     {
-        switch (varItem->getDataType())
-        {
-        case typePointer:
-            queueCmd(new GDBPointerCommand(varItem));
-            break;
-
-        default:
-             //rgr: we need to do this in order to move the output modifier
-             //     from the middle to the beginning
-             QString strCmd = varItem->fullName();
-             int iFound = strCmd.find( QRegExp("./[xd] ", FALSE) );
-             if (iFound != -1) {
-               strCmd.prepend( strCmd.mid(iFound+1, 3) );
-               strCmd.replace( QRegExp("./[xd] "), "." );
-             }
-             queueCmd(new GDBItemCommand(varItem, QCString("print ") + strCmd.latin1()));            break;
+        //rgr: we need to do this in order to move the output modifier
+        //     from the middle to the beginning
+        QString strCmd = varItem->gdbExpression();
+        int iFound = strCmd.find( QRegExp("./[xd] ", FALSE) );
+        if (iFound != -1) {
+            strCmd.prepend( strCmd.mid(iFound+1, 3) );
+            strCmd.replace( QRegExp("./[xd] "), "." );
         }
+        queueCmd(new GDBItemCommand(varItem, QCString("print ") + strCmd.latin1()));
         return;
     }
 
@@ -2023,11 +2128,11 @@ void GDBController::slotDbgWroteStdin(KProcess *)
 
 void GDBController::slotDbgProcessExited(KProcess* process)
 {
-    if ( process->exitStatus() == 127 )
-      emit debuggerRunError(127);
+    if ( !process->normalExit() )
+        emit debuggerAbnormalExit();
 
     destroyCmds();
-    state_ = s_appNotStarted|s_programExited|(state_&(s_viewLocals|s_shuttingDown));
+    state_ = s_dbgNotStarted|s_appNotStarted|s_programExited|(state_&(s_viewLocals|s_shuttingDown));
     emit dbgStatus (i18n("Process exited"), state_);
 
     emit gdbStdout("(gdb) Process exited\n");
